# mypy: disable-error-code="no-untyped-def, no-untyped-call, assignment"
import json
import locale
import logging
import sys
import tempfile
from pathlib import Path
from textwrap import dedent
from typing import cast

import pytest
from click.testing import CliRunner

from pyproject2conda.cli import app

ROOT = Path(__file__).resolve().parent / "data"


def do_run(runner, command, *opts, filename=None, must_exist=False, **kwargs):
    if filename is None:
        raise ValueError
    filename = Path(filename)
    if must_exist and not filename.exists():
        msg = f"filename {filename} does not exist"
        raise ValueError(msg)

    return runner.invoke(app, [command, "-f", str(filename), *opts], **kwargs)


def check_result(result, expected) -> None:
    assert result.output == dedent(expected)


@pytest.fixture(params=["test-pyproject.toml"])
def filename(request) -> Path:
    return ROOT / cast(str, request.param)


@pytest.mark.parametrize(
    ("fname", "expected"),
    [
        (
            "test-pyproject.toml",
            """\
            Extras
            ======
            * build-system.requires
            * dev
            * dev-extras
            * dist-pypi
            * test
            Groups
            ======
            * build-system.requires
            """,
        ),
        (
            "test-pyproject-groups.toml",
            """\
            Extras
            ======
            * build-system.requires
            Groups
            ======
            * build-system.requires
            * dev
            * dev-extras
            * dist-pypi
            * test
            """,
        ),
    ],
)
def test_list(fname: str, expected: str, runner: CliRunner) -> None:
    filename = ROOT / fname
    for cmd in ["l", "list"]:
        result = do_run(runner, cmd, filename=filename)
        check_result(result, expected)

    result = do_run(runner, "list", "-v", filename=filename)
    check_result(result, expected)


def test_no_file(runner) -> None:
    result = do_run(runner, "yaml", filename="hello/there.toml")
    assert isinstance(result.exception, FileNotFoundError)


@pytest.mark.parametrize(
    ("fname", "style"),
    [
        ("test-pyproject.toml", "extras"),
        ("test-pyproject-groups.toml", "groups"),
    ],
)
def test_create(fname, style, runner) -> None:
    # test unknown file
    filename = ROOT / fname
    extra_or_group_opts = ["-e", "--extra"] if style == "extras" else ["-g", "--group"]

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    result = do_run(runner, "yaml", filename=filename)

    check_result(result, expected)

    cmd = " ".join([Path(sys.argv[0]).name] + sys.argv[1:])

    expected = f"""\
#
# This file is autogenerated by pyproject2conda
# with the following command:
#
#     $ {cmd}
#
# You should not manually edit this file.
# Instead edit the corresponding pyproject.toml file.
#
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    result = do_run(runner, "yaml", "--header", filename=filename)

    check_result(result, expected)

    # -p flag
    expected = """\
channels:
  - conda-forge
dependencies:
  - python>=3.8,<3.11
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    for opt in [("--python-include", "infer")]:
        for cmd in ["y", "yaml"]:
            result = do_run(runner, cmd, *opt, filename=filename)
            check_result(result, expected)

    # -p python=3.8
    expected = """\
channels:
  - conda-forge
dependencies:
  - python=3.8
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """
    for opts in [
        ("--python-include", "python=3.8"),
        ("--python", "3.8"),
        ("-p", "3.8"),
    ]:
        result = do_run(runner, "yaml", *opts, filename=filename)
        check_result(result, expected)

    # dev
    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - additional-thing
  - conda-matplotlib
  - pip
  - pip:
      - athing
    """

    for opt in extra_or_group_opts:
        result = do_run(runner, "yaml", opt, "dev", "--no-sort", filename=filename)
        check_result(result, expected)

    # test if add in "test" gives same answer
    opts = extra_or_group_opts[0]
    result = do_run(
        runner, "yaml", opt, "dev", opt, "test", "--no-sort", filename=filename
    )
    check_result(result, expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - additional-thing
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - conda-matplotlib
  - pandas
  - pip
  - pip:
      - athing
    """

    for opt in extra_or_group_opts:
        result = do_run(runner, "yaml", opt, "dev", filename=filename)
        check_result(result, expected)

    # test if add in "test" gives same answer
    opt = extra_or_group_opts[0]
    result = do_run(runner, "yaml", opt, "dev", opt, "test", filename=filename)
    check_result(result, expected)

    # override channel
    expected = """\
channels:
  - hello
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    for opt in ["-c", "--channel"]:
        result = do_run(runner, "yaml", opt, "hello", filename=filename)
        check_result(result, expected)

    expected = """\
channels:
  - hello
  - there
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pip
  - pip:
      - athing
    """

    for opt in ["-c", "--channel"]:
        result = do_run(runner, "yaml", opt, "hello", opt, "there", filename=filename)
        check_result(result, expected)

    # test
    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - pandas
  - conda-forge::pytest
  - pip
  - pip:
      - athing
    """

    opt = extra_or_group_opts[0]
    result = do_run(runner, "yaml", opt, "test", "--no-sort", filename=filename)
    check_result(result, expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - pandas
  - pip
  - pip:
      - athing
    """
    result = do_run(runner, "yaml", opt, "test", filename=filename)
    check_result(result, expected)

    # isolated
    expected = """\
channels:
  - conda-forge
dependencies:
  - setuptools
  - pip
  - pip:
      - build
    """
    for opt in extra_or_group_opts:
        result = do_run(
            runner, "yaml", opt, "dist-pypi", "--skip-package", filename=filename
        )
        check_result(result, expected)


@pytest.mark.parametrize(
    ("fname", "style"),
    [
        ("test-pyproject-reorder.toml", "extras"),
        ("test-pyproject-reorder.toml", "groups"),
    ],
)
def test_create_reorder(fname, style, runner) -> None:
    filename = ROOT / fname
    extra_or_group_opts = ["-e", "--extra"] if style == "extras" else ["-g", "--group"]

    # different ordering
    expected = """\
channels:
  - conda-forge
dependencies:
  - conda-forge::cthing
  - bthing-conda
  - conda-forge::pytest
  - pandas
  - conda-matplotlib
  - additional-thing
  - pip
  - pip:
      - athing
    """

    for opt in extra_or_group_opts:
        result = do_run(
            runner,
            "yaml",
            opt,
            "dev",
            "--no-sort",
            filename=filename,
            must_exist=True,
        )
        check_result(result, expected)

    # test if add in "test" gives same answer
    opt = extra_or_group_opts[0]
    result = do_run(
        runner,
        "yaml",
        opt,
        "dev",
        opt,
        "test",
        "--no-sort",
        filename=filename,
        must_exist=True,
    )
    check_result(result, expected)

    expected = """\
channels:
  - conda-forge
dependencies:
  - additional-thing
  - bthing-conda
  - conda-forge::cthing
  - conda-forge::pytest
  - conda-matplotlib
  - pandas
  - pip
  - pip:
      - athing
    """

    for opt in extra_or_group_opts:
        result = do_run(
            runner,
            "yaml",
            opt,
            "dev",
            filename=filename,
            must_exist=True,
        )
        check_result(result, expected)

    # test if add in "test" gives same answer
    opt = extra_or_group_opts[0]
    result = do_run(
        runner,
        "yaml",
        opt,
        "dev",
        opt,
        "test",
        filename=filename,
        must_exist=True,
    )
    check_result(result, expected)


@pytest.mark.parametrize(
    ("fname", "opt"),
    [
        ("test-pyproject.toml", "-e"),
        ("test-pyproject-groups.toml", "-g"),
    ],
)
def test_requirements(fname, opt, runner) -> None:
    filename = ROOT / fname

    expected = """\
athing
bthing
cthing; python_version < "3.10"
    """

    for cmd in ["r", "requirements"]:
        result = do_run(runner, cmd, filename=filename)
        check_result(result, expected)

    expected = """\
athing
bthing
cthing; python_version < "3.10"
pandas
pytest
matplotlib
    """

    result = do_run(runner, "requirements", opt, "dev", "--no-sort", filename=filename)
    check_result(result, expected)

    result = do_run(
        runner,
        "requirements",
        opt,
        "dev",
        opt,
        "test",
        "--no-sort",
        filename=filename,
    )
    check_result(result, expected)

    expected = """\
athing
bthing
cthing; python_version < "3.10"
pandas
pytest
matplotlib
thing; python_version < "3.10"
other
    """

    result = do_run(
        runner,
        "requirements",
        opt,
        "dev",
        "--no-sort",
        "-r",
        "thing;python_version<'3.10'",
        "-r",
        "other",
        filename=filename,
    )

    check_result(result, expected)

    expected = """\
athing
bthing
cthing; python_version < "3.10"
matplotlib
pandas
pytest
    """

    result = do_run(runner, "requirements", opt, "dev", filename=filename)
    check_result(result, expected)

    result = do_run(runner, "requirements", opt, "dev", opt, "test", filename=filename)
    check_result(result, expected)

    expected = """\
athing
bthing
cthing; python_version < "3.10"
matplotlib
other
pandas
pytest
thing; python_version < "3.10"
    """

    result = do_run(
        runner,
        "requirements",
        opt,
        "dev",
        "-r",
        "thing;python_version<'3.10'",
        "-r",
        "other",
        filename=filename,
    )

    check_result(result, expected)

    # allow whitespace:
    expected = """\
athing
bthing
cthing; python_version < "3.10"
matplotlib
other
pandas
pytest
thing; python_version < "3.10"
    """

    result = do_run(
        runner,
        "requirements",
        opt,
        "dev",
        "-r",
        "thing; python_version < '3.10'",
        "-r",
        "other",
        "--no-remove-whitespace",
        filename=filename,
    )

    check_result(result, expected)

    expected = """\
setuptools
build
    """

    result = do_run(
        runner,
        "requirements",
        opt,
        "dist-pypi",
        "--skip-package",
        "--no-sort",
        filename=filename,
    )
    check_result(result, expected)

    expected = """\
build
setuptools
    """

    result = do_run(
        runner, "requirements", opt, "dist-pypi", "--skip-package", filename=filename
    )
    check_result(result, expected)


def check_results_conda_req(path, expected) -> None:
    with Path(path).open(encoding=locale.getpreferredencoding(False)) as f:
        result = f.read()

    assert result == dedent(expected)


def test_conda_requirements(filename, runner) -> None:
    result = do_run(runner, "c", "hello.txt", filename=filename)

    assert isinstance(result.exception, ValueError)

    result = do_run(runner, "c", "--prefix", "hello", "a", "b", filename=filename)

    assert isinstance(result.exception, ValueError)

    # stdout
    result = do_run(runner, "c", filename=filename)

    expected = """\
#conda requirements
bthing-conda
conda-forge::cthing
pip

#pip requirements
athing
    """

    assert result.output == dedent(expected)

    with tempfile.TemporaryDirectory() as d_tmp:
        d = Path(d_tmp)

        expected_conda = """\
bthing-conda
conda-forge::cthing
pip
        """
        expected_pip = """\
athing
        """

        for cmd in ["c", "conda-requirements"]:
            do_run(
                runner,
                cmd,
                "--prefix",
                str(d / "hello-"),
                "--no-header",
                filename=filename,
            )

            check_results_conda_req(d / "hello-conda.txt", expected_conda)
            check_results_conda_req(d / "hello-pip.txt", expected_pip)

        do_run(
            runner,
            "conda-requirements",
            str(d / "conda-output.txt"),
            str(d / "pip-output.txt"),
            "--no-header",
            filename=filename,
        )

        check_results_conda_req(d / "conda-output.txt", expected_conda)
        check_results_conda_req(d / "pip-output.txt", expected_pip)

        do_run(
            runner,
            "c",
            "--prefix",
            str(d / "hello-"),
            "--prepend-channel",
            "-c",
            "achannel",
            "--no-header",
            filename=filename,
        )

        expected_conda = """\
achannel::bthing-conda
conda-forge::cthing
achannel::pip
        """

        check_results_conda_req(d / "hello-conda.txt", expected_conda)
        check_results_conda_req(d / "hello-pip.txt", expected_pip)


def check_results_json(path, expected) -> None:
    with Path(path).open(encoding=locale.getpreferredencoding(False)) as f:
        result = json.load(f)

    assert result == expected


@pytest.mark.parametrize(
    ("fname", "opt"),
    [
        ("test-pyproject.toml", "-e"),
        ("test-pyproject-groups.toml", "-g"),
    ],
)
def test_json(fname, opt, runner) -> None:
    filename = ROOT / fname
    # stdout
    result = do_run(runner, "j", filename=filename)

    expected = """\
{"dependencies": ["bthing-conda", "conda-forge::cthing", "pip"], "pip": ["athing"], "channels": ["conda-forge"]}
    """

    assert result.output == dedent(expected)

    with tempfile.TemporaryDirectory() as d_tmp:
        d = Path(d_tmp)

        expected = {
            "dependencies": ["bthing-conda", "conda-forge::cthing", "pip"],
            "pip": ["athing"],
            "channels": ["conda-forge"],
        }

        path = d / "hello.json"

        do_run(runner, "json", "-o", str(path), filename=filename)

        check_results_json(d / "hello.json", expected)

        orig_time = path.stat().st_mtime
        do_run(runner, "json", "-o", str(path), filename=filename)
        check_results_json(d / "hello.json", expected)

        assert path.stat().st_mtime == orig_time

        expected = {
            "dependencies": [
                "bthing-conda",
                "conda-forge::cthing",
                "pandas",
                "conda-forge::pytest",
                "additional-thing",
                "conda-matplotlib",
                "pip",
            ],
            "pip": ["athing"],
            "channels": ["conda-forge"],
        }

        do_run(
            runner,
            "json",
            "-o",
            str(d / "there.json"),
            opt,
            "dev",
            "--no-sort",
            filename=filename,
        )

        check_results_json(d / "there.json", expected)

        expected = {
            "dependencies": [
                "additional-thing",
                "bthing-conda",
                "conda-forge::cthing",
                "conda-forge::pytest",
                "conda-matplotlib",
                "pandas",
                "pip",
            ],
            "pip": ["athing"],
            "channels": ["conda-forge"],
        }

        do_run(
            runner,
            "json",
            "-o",
            str(d / "there.json"),
            opt,
            "dev",
            "-w",
            "force",
            filename=filename,
        )

        check_results_json(d / "there.json", expected)


def test_json_no_channel(runner) -> None:
    filename = ROOT / "simple-pyproject.toml"
    expected = {
        "dependencies": ["bthing-conda", "conda-forge::cthing", "pip"],
        "pip": ["athing"],
    }

    with tempfile.TemporaryDirectory() as d_tmp:
        d = Path(d_tmp)
        do_run(
            runner,
            "json",
            "-o",
            str(d / "there.json"),
            filename=filename,
        )

        check_results_json(d / "there.json", expected)


def test_alias(filename, runner) -> None:
    result = do_run(runner, "q", filename=filename)

    assert isinstance(result.exception, BaseException)


def test_overwrite(filename, caplog) -> None:
    runner = CliRunner(mix_stderr=True)

    caplog.set_level(logging.INFO)

    with tempfile.TemporaryDirectory() as d_tmp:
        d = Path(d_tmp)

        path = d / "out.yaml"

        assert not path.exists()

        do_run(
            runner,
            "yaml",
            "-o",
            str(path),
            "-v",
            "-w",
            "force",
            catch_exceptions=False,
            filename=filename,
        )
        assert f"Creating yaml {d_tmp}/out.yaml" in caplog.text

        orig_time = path.stat().st_mtime

        for cmd in ["check", "skip", "force"]:
            do_run(
                runner,
                "yaml",
                "-o",
                str(d / "out.yaml"),
                "-v",
                "-w",
                cmd,
                catch_exceptions=False,
                filename=filename,
            )

            if cmd == "force":
                assert path.stat().st_mtime > orig_time
            else:
                assert path.stat().st_mtime == orig_time

            assert (
                f"Skipping yaml {d_tmp}/out.yaml. Pass `-w force` to force recreate output"
                in caplog.text
            )

        path = d / "out.txt"
        assert not path.exists()

        do_run(
            runner,
            "r",
            "-o",
            str(d / "out.txt"),
            "-v",
            "-w",
            "force",
            catch_exceptions=False,
            filename=filename,
        )
        orig_time = path.stat().st_mtime

        assert f"Creating requirements {d_tmp}/out.txt" in caplog.text

        for cmd in ["check", "skip", "force"]:
            do_run(
                runner,
                "r",
                "-o",
                str(path),
                "-v",
                "-w",
                cmd,
                catch_exceptions=False,
                filename=filename,
            )

            if cmd == "force":
                assert path.stat().st_mtime > orig_time
            else:
                assert path.stat().st_mtime == orig_time
            assert (
                f"Skipping requirements {d_tmp}/out.txt. Pass `-w force` to force recreate output"
                in caplog.text
            )


@pytest.mark.parametrize("fname", ["test-pyproject.toml", "test-pyproject-groups.toml"])
def test_userconfig(fname, runner) -> None:
    filename = ROOT / fname
    expected = """\
# --------------------
# Creating yaml py310-user-dev.yaml
name: hello-there
channels:
  - conda-forge
dependencies:
  - python=3.10
  - bthing-conda
  - conda-forge::pytest
  - conda-matplotlib
  - pandas
  - setuptools
  - pip
  - pip:
      - athing
      - build
    """

    results = do_run(
        runner,
        "p",
        "--dry",
        "--envs",
        "user-dev",
        "--user-config",
        str(ROOT / "config" / "userconfig2.toml"),
        filename=filename,
    )

    check_result(results, expected)



"""
Utility methods (:mod:`pyproject2conda.utils`)
==============================================
"""

from __future__ import annotations

import enum
import re
from pathlib import Path
from typing import TYPE_CHECKING

from packaging.version import Version

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable, Mapping, Sequence

    from ._typing import T


# taken from https://github.com/python-attrs/attrs/blob/main/src/attr/_make.py
class _Missing(enum.Enum):
    """
    Sentinel to indicate the lack of a value when ``None`` is ambiguous.

    If extending attrs, you can use ``typing.Literal[MISSING]`` to show
    that a value may be ``MISSING``.

    .. versionchanged:: 21.1.0 ``bool(MISSING)`` is now False.
    .. versionchanged:: 22.2.0 ``MISSING`` is now an ``enum.Enum`` variant.
    """

    MISSING = enum.auto()

    def __repr__(self) -> str:
        return "MISSING"  # pragma: no cover

    def __bool__(self) -> bool:
        return False  # pragma: no cover


MISSING = _Missing.MISSING
"""
Sentinel to indicate the lack of a value when ``None`` is ambiguous.
"""


# taken from https://github.com/conda/conda-lock/blob/main/conda_lock/common.py
def get_in(
    keys: Sequence[Any],
    nested_dict: Mapping[Any, Any],
    default: Any = None,
    factory: Callable[[], Any] | None = None,
) -> Any:
    """
    >>> foo = {"a": {"b": {"c": 1}}}
    >>> get_in(["a", "b"], foo)
    {'c': 1}

    """
    import operator
    from functools import reduce

    try:
        return reduce(operator.getitem, keys, nested_dict)
    except (KeyError, IndexError, TypeError):
        if factory is not None:
            return factory()
        return default


def get_default_pythons(path: str | Path = ".python-version") -> list[str]:
    """Get default python value from .python-version file"""
    path = Path(path)
    if path.exists():
        out = path.read_text().split()
        # only keep major.minor
        out[0] = ".".join(out[0].split(".")[:2])
        return out
    return []


def get_all_pythons(
    data: dict[str, Any] | None, path: str | Path | None = None
) -> list[str]:
    """Get python versions from pyproject:project.classifiers"""
    if data is None:
        assert path is not None  # noqa: S101

        from ._compat import tomllib

        with Path(path).open("rb") as f:
            data = tomllib.load(f)

    return [
        c.split()[-1]
        for c in get_in(["project", "classifiers"], data, factory=list)
        if c.startswith("Programming Language :: Python :: 3.")
    ]


def get_lowest_version(versions: Iterable[str]) -> str:
    """Get lowest version"""
    return min(versions, key=Version)


def get_highest_version(versions: Iterable[str]) -> str:
    """Get highest version"""
    return max(versions, key=Version)


def select_pythons(
    pythons: Sequence[str],
    default_pythons: list[str],
    all_pythons: list[str],
) -> list[str]:
    """Select pythons from string values."""
    if len(pythons) == 1:
        python = pythons[0]

        if python == "default":
            if not default_pythons:
                msg = "Must include `.python-version` to use `python = 'default'`."
                raise ValueError(msg)
            return default_pythons

        if python in {"all", "lowest", "highest"}:
            if not all_pythons:
                msg = "Must specify python versions in project.classifiers table to use `python` in `{'all', 'lowest', 'highest'}`"
                raise ValueError(msg)
            return (
                all_pythons
                if python == "all"
                else [get_lowest_version(all_pythons)]
                if python == "lowest"
                else [get_highest_version(all_pythons)]
            )

    return list(pythons)


def parse_pythons(
    python_include: str | None,
    python_version: str | None,
    python: str | None,
    toml_path: str | Path,
) -> tuple[str | None, str | None]:
    """Create python_include/python_version."""
    if python:
        python = select_pythons(
            [python],
            get_default_pythons(),
            get_all_pythons(data=None, path=toml_path),
        )[0]
        return f"python={python}", python

    return python_include, python_version


def update_target(
    target: str | Path | None,
    *deps: str | Path,
    overwrite: str = "check",
) -> bool:
    """Check if target is older than deps:"""
    if target is None:
        # No output file. always run.
        return True

    overwrite = overwrite.lower()
    target = Path(target)

    if overwrite == "force":
        update = True
    elif overwrite == "skip":
        update = not target.exists()

    elif overwrite == "check":
        if not target.exists():
            update = True
        else:
            deps_filtered: list[Path] = [d for d in map(Path, deps) if d.exists()]

            target_time = target.stat().st_mtime

            update = any(target_time < dep.stat().st_mtime for dep in deps_filtered)
    else:  # pragma: no cover
        msg = f"unknown option overwrite={overwrite}"
        raise ValueError(msg)

    return update


# * filename from template
def filename_from_template(
    template: str | None,
    python: str | None = None,
    python_version: str | None = None,
    env_name: str | None = None,
    ext: str | None = ".yaml",
) -> str | None:
    """
    Create a filename from

    --python-include python=3.8 or --python-version 3.8 or --python 3.8
    py_version: 3.8
    py: 38

    env : name of environment
    """
    if template is None:
        return None

    kws: dict[str, str] = {}
    if python:
        py_version = python
    elif python_version:
        py_version = python_version
    else:
        py_version = None

    if py_version:
        kws["py_version"] = py_version
        kws["py"] = py_version.replace(".", "")

    if env_name:  # pragma: no cover
        kws["env"] = env_name

    if ext:  # pragma: no cover
        template += f"{ext}"

    return template.format(**kws)


_WHITE_SPACE_REGEX = re.compile(r"\s+")


def remove_whitespace(s: str) -> str:
    """Cleanup whitespace from string."""
    return re.sub(_WHITE_SPACE_REGEX, "", s)


def remove_whitespace_list(s: Iterable[str]) -> list[str]:
    """Cleanup whitespace from list of strings."""
    return [remove_whitespace(x) for x in s]


def unique_list(values: Iterable[T]) -> list[T]:
    """
    Return only unique values in list.
    Unlike using set(values), this preserves order.
    """
    output: list[T] = []
    for v in values:
        if v not in output:
            output.append(v)
    return output


def list_to_str(values: Iterable[str] | None, eol: bool = True) -> str:
    """Join list of strings with newlines to single string."""
    if values:
        output = "\n".join(values)
        if eol:
            output += "\n"
    else:
        output = ""

    return output



# pyright: reportUnknownMemberType=false
"""
Console script for pyproject2conda (:mod:`~pyproject2conda.cli`)
================================================================
"""
# * Imports -------------------------------------------------------------------

import locale
import logging
import os
from enum import Enum
from functools import lru_cache, wraps
from inspect import signature
from pathlib import Path
from typing import Any, Callable, List, Optional, Union, cast

# from click import click.Context
import click
import typer
from typer.core import TyperGroup

from pyproject2conda import __version__
from pyproject2conda.requirements import ParseDepends
from pyproject2conda.utils import (
    parse_pythons,
    update_target,
)

from ._typing import R
from ._typing_compat import Annotated

# * Logger -----------------------------------------------------------------------------

FORMAT = "%(message)s [%(name)s - %(levelname)s]"
logging.basicConfig(level=logging.WARNING, format=FORMAT)
logger = logging.getLogger("pyproject2conda")


# * Settings ---------------------------------------------------------------------------
if "P2C_COLUMNS" in os.environ:
    os.environ["COLUMNS"] = os.environ["P2C_COLUMNS"]  # pragma: no cover


# * Typer App --------------------------------------------------------------------------


class AliasedGroup(TyperGroup):
    """Provide aliasing for commands"""

    def get_command(
        self, ctx: click.Context, cmd_name: str
    ) -> Optional[click.core.Command]:
        rv = super().get_command(ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        if len(matches) == 1:
            return super().get_command(ctx, matches[0])
        ctx.fail(
            "Too many matches: {}".format(", ".join(sorted(matches)))
        )  # pragma: no cover
        return None  # pragma: no cover

    def list_commands(self, ctx: click.Context) -> List[str]:  # noqa: ARG002
        return list(self.commands)


app_typer = typer.Typer(cls=AliasedGroup, no_args_is_help=True)


def version_callback(value: bool) -> None:
    """Versioning call back."""
    if value:
        typer.echo(f"pyproject2conda, version {__version__}")
        raise typer.Exit


@app_typer.callback()
def main(
    version: bool = typer.Option(
        None, "--version", "-v", callback=version_callback, is_eager=True
    ),
) -> None:
    """
    Extract conda `environment.yaml` and pip `requirement.txt` files from `pyproject.toml`

    Note that all subcommands can be called with shortest possible match. Also, you can
    call with any of `pyproject2conda`, `p2c`, `python -m pyproject2conda`.
    For example,

    .. code-block:: console

            # these are equivalent
            $ pyproject2conda yaml ...
            $ p2c y ...
            $ python -m pyproject2conda ...
    """
    return


# * Options ----------------------------------------------------------------------------

DEFAULT_TOML_PATH = Path("./pyproject.toml")

PYPROJECT_CLI = Annotated[
    Path,
    typer.Option(  # pyright: ignore[reportUnknownMemberType]
        "--file",
        "-f",
        help="input pyproject.toml file",
    ),
]
EXTRAS_CLI = Annotated[
    Optional[List[str]],
    typer.Option(  # pyright: ignore[reportUnknownMemberType]
        "--extra",
        "-e",
        help="""
        Include dependencies from extra <extra> from `project.optional-dependencies` table of `pyproject.toml`.
        Can specify multiple times for multiple extras.
        Use name `extras` for specifying in `pyproject.toml`
        """,
    ),
]
GROUPS_CLI = Annotated[
    Optional[List[str]],
    typer.Option(  # pyright: ignore[reportUnknownMemberType]
        "--group",
        "-g",
        help="""
        Include dependencies from group <group> from `dependency-groups` table of `pyproject.toml`.
        Can specify multiple times for multiple groups.
        Use name `groups` for specifying in `pyproject.toml`
        """,
    ),
]
EXTRAS_OR_GROUPS_CLI = Annotated[
    Optional[List[str]],
    typer.Option(  # pyright: ignore[reportUnknownMemberType]
        "--extra-or-group",
        help="""
        Include dependencies from extra or group of `pyproject.toml`.
        Extras are checked first, followed by groups.  The first instance of `extra-or-group` found is used.
        That is, if both `extras` and `groups` contain `extra-or-group`, the extra will be used.
        Use name `extras-or-groups` for specifying in `pyproject.toml`
        """,
    ),
]


CHANNEL_CLI = Annotated[
    Optional[List[str]],
    typer.Option(  # pyright: ignore[reportUnknownMemberType]
        "--channel",
        "-c",
        help="Conda channel.  Can specify. Overrides [tool.pyproject2conda.channels]",
    ),
]
NAME_CLI = Annotated[
    Optional[str],
    typer.Option(  # pyright: ignore[reportUnknownMemberType]
        "--name",
        "-n",
        help="Name of conda env",
    ),
]
OUTPUT_CLI = Annotated[
    Optional[Path],
    typer.Option(  # pyright: ignore[reportUnknownMemberType]
        "--output",
        "-o",
        help="File to output results",
    ),
]


class Overwrite(str, Enum):
    """Options for `--overwrite`"""

    check = "check"
    skip = "skip"
    foce = "force"


OVERWRITE_CLI = Annotated[
    Overwrite,
    typer.Option(
        "--overwrite",
        "-w",
        case_sensitive=False,
        help="""
    What to do if output file exists.
    (check): Create if missing. If output exists and passed `--filename` is newer, recreate output, else skip.
    (skip): If output exists, skip.
    (force): force: force recreate output.
    """,
    ),
]

VERBOSE_CLI = Annotated[
    Optional[int],
    typer.Option(
        "--verbose",
        "-v",
        help="Pass `-v/--verbose` for verbose output.  Pass multiple times to set verbosity level.",
        count=True,
    ),
]
SKIP_PACKAGE_CLI = Annotated[
    bool,
    typer.Option(
        "--skip-package",
        help="""
        Default is to include package dependencies from `project.dependencies`
        table of `pyproject.toml`. Passing `--skip-package` (or `skip_package =
        true` in `tool.pyproject2conda.envs...` table of `pyproject.toml`) will
        exclude the package dependencies. This is useful to define environments
        that should exclude base dependencies (like build, etc) in
        pyproject.toml.
        """,
    ),
]
SORT_DEPENDENCIES_CLI = Annotated[
    bool,
    typer.Option(
        "--sort/--no-sort",
        help="""
        Default is to sort the dependencies (excluding `--python-include`).
        Pass `--no-sort` to instead place dependencies in order they are
        gathered.
        """,
    ),
]
PYTHON_INCLUDE_CLI = Annotated[
    Optional[str],
    typer.Option(
        "--python-include",
        help="""
        If value passed, use this value (exactly) in the output. So, for
        example, pass `--python-include "python=3.8"`. Special case is the
        value `"infer"`. This infers the value of python from `pyproject.toml`
        """,
    ),
]
PYTHON_VERSION_CLI = Annotated[
    Optional[str],
    typer.Option(
        "--python-version",
        help="""
         Python version to check `python_version <=> {python_version}` lines
         against. That is, this version is used to limit packages in resulting
         output. For example, if have a line like `a-package; python_version <
         '3.9'`, Using `--python-version 3.10` will not include `a-package`,
         while `--python-version 3.8` will include `a-package`.
         """,
    ),
]
PYTHON_CLI = Annotated[
    Optional[str],
    typer.Option(
        "--python",
        "-p",
        help="""
        python version. passing `--python {version}` is equivalent to passing
        `--python-version={version} --python-include=python{version}`. if
        passed, this overrides values of passed via `--python-version` and
        `--python-include`. pass `--python="default"` to include the python
        version (major.minor only) from `.python-version` file in the current
        directory. pass `"lowest"` or `"highest"` to include the lowest or
        highest python version, respectively, from
        `pyproject.toml:project.classifiers` table. in project mode, you can
        pass multiple python version in `pyproject.toml` with, e.g., `python =
        ["3.8", "3.9", ....]`, or using `python = "all"`, to include all python
        versions extracted from `pyproject.toml:project.classifiers` table.
        """,
    ),
]
HEADER_CLI = Annotated[
    Optional[bool],
    typer.Option(
        "--header/--no-header",
        help="""
        If True (--header) include header line in output. Default is to include
        the header for output to a file, and not to include header when writing
        to stdout.
        """,
    ),
]
DEPS_CLI = Annotated[
    Optional[List[str]],
    typer.Option(
        "--deps",
        "-d",
        help="Additional conda dependencies.",
    ),
]
REQS_CLI = Annotated[
    Optional[List[str]],
    typer.Option(
        "--reqs",
        "-r",
        help="""
        Additional pip requirements. For example, pass `-r '-e .'` to included
        editable version of current package in requirements file.
        """,
    ),
]
ENVS_CLI = Annotated[
    Optional[List[str]],
    typer.Option(
        help="List of environments to build files for.  Default to building all environments",
    ),
]
TEMPLATE_CLI = Annotated[
    Optional[str],
    typer.Option(
        help="Template for environments that do not have a python version. Defaults to `{env}`."
    ),
]
TEMPLATE_PYTHON_CLI = Annotated[
    Optional[str],
    typer.Option(
        help="""
        Template for environments that do have a python version. Defaults to
        "py{py}-{env}". For example, with `--template-python="py{py}-{env}"` and
        `--python=3.8` and environment "dev", output would be "py38-dev"
        \b
        * {py} -> "38"
        * {py_version} -> "3.8"
        * {env} -> "dev"
        """,
    ),
]
REQS_EXT_CLI = Annotated[
    Optional[str],
    typer.Option(
        "--reqs-ext",
        help="""
        Extension to use with requirements file output created from template.  Defaults to `".txt"`.
        """,
    ),
]
YAML_EXT_CLI = Annotated[
    Optional[str],
    typer.Option(
        "--yaml-ext",
        help="""
        Extension to use with conda environment.yaml file output created from template.  Defaults to `".yaml"`
        """,
    ),
]
DRY_CLI = Annotated[
    bool,
    typer.Option(
        "--dry",
        help="If passed, do a dry run",
    ),
]
USER_CONFIG_CLI = Annotated[
    Optional[str],
    typer.Option(
        "--user-config",
        help="""
        Additional toml file to supply configuration. This can be used to
        override/add environment files for your own use (apart from project env
        files). The (default) value `infer` means to infer the configuration
        from `--filename`.
        """,
    ),
]
# For conda-requirements
PREFIX_CLI = Annotated[
    Optional[str],
    typer.Option(
        "--prefix",
        help="set conda-output=prefix + 'conda.txt', pip-output=prefix + 'pip.txt'",
    ),
]
PREPEND_CHANNEL_CLI = Annotated[
    bool,
    typer.Option(
        "--prepend-channel",
    ),
]


ALLOW_EMPTY_OPTION = typer.Option(
    "--allow-empty/--no-allow-empty",
    help="""
    What to do if there are no package requirements for an environment. The
    default (`--no-allow-empty`) is to raise an error if the specification
    leads to no requirements. Passing `--allow-empty` will lead to a message
    being printed, but no environment file being created.
    """,
)


REMOVE_WHITESPACE_OPTION = typer.Option(
    "--remove-whitespace/--no-remove-whitespace",
    help="""
    What to do with whitespace in a dependency. Passing `--remove-whitespace`
    will remove whitespace in a given dependency. For example, the dependency
    `package >= 1.0` will be converted to `package>=1.0`. Pass
    `--no-remove-whitespace` to keep the the whitespace in the output.
    """,
)


# * Utils ------------------------------------------------------------------------------
def _get_header_cmd(
    header: Optional[bool], output: Union[str, Path, None]
) -> Optional[str]:
    if header is None:
        header = output is not None

    if header:
        import sys
        from pathlib import Path

        return " ".join([Path(sys.argv[0]).name] + sys.argv[1:])

    return None


@lru_cache
def _get_requirement_parser(filename: Union[str, Path]) -> ParseDepends:
    return ParseDepends.from_path(filename)


def _log_skipping(
    logger: logging.Logger, style: str, output: Union[str, Path, None]
) -> None:
    logger.info(
        "Skipping %s %s. Pass `-w force` to force recreate output", style, output
    )


def _log_creating(
    logger: logging.Logger,
    style: str,
    output: Union[str, Path, None],
    prefix: Optional[str] = None,
) -> None:
    if prefix is None:  # pragma: no cover
        prefix = "# " if prefix is None and output is None else ""

    s = f"{prefix}Creating {style}"

    if output:
        s = f"{s} {output}"

    logger.info(s)


def add_verbose_logger(
    logger: logging.Logger, verbose_arg: str = "verbose"
) -> Callable[[Callable[..., R]], Callable[..., R]]:
    """Decorator factory to add logger and set logger level based on verbosity argument value."""

    def decorator(func: Callable[..., R]) -> Callable[..., R]:
        bind = signature(func).bind

        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> R:
            params = bind(*args, **kwargs)
            params.apply_defaults()

            verbosity = cast("Optional[int]", params.arguments[verbose_arg])

            if verbosity is None:
                # leave where it is:
                pass
            else:
                if verbosity < 0:  # pragma: no cover
                    level = logging.ERROR
                elif verbosity == 0:  # pragma: no cover
                    level = logging.WARNING
                elif verbosity == 1:
                    level = logging.INFO
                else:  # pragma: no cover
                    level = logging.DEBUG

                logger.setLevel(level)

            # add error logger to function call
            try:
                return func(*args, **kwargs)
            except Exception:
                logger.exception("found error")
                raise

        return wrapped

    return decorator


# * Commands ---------------------------------------------------------------------------
# ** List
# @app_typer.command("l", hidden=True)
@app_typer.command("list")
@add_verbose_logger(logger)
def create_list(
    filename: PYPROJECT_CLI = DEFAULT_TOML_PATH,
    verbose: VERBOSE_CLI = None,
) -> None:
    """List available extras."""
    logger.info("filename: %s", filename)

    d = _get_requirement_parser(filename)

    for name, vals in [("Extras", d.extras), ("Groups", d.groups)]:
        print(name)
        print("======")
        for val in sorted(vals):
            print("*", val)


# ** Yaml
# @app_typer.command("y", hidden=True)
@app_typer.command()
@add_verbose_logger(logger)
def yaml(
    filename: PYPROJECT_CLI = DEFAULT_TOML_PATH,
    extras: EXTRAS_CLI = None,
    groups: GROUPS_CLI = None,
    extras_or_groups: EXTRAS_OR_GROUPS_CLI = None,
    channels: CHANNEL_CLI = None,
    output: OUTPUT_CLI = None,
    name: NAME_CLI = None,
    python_include: PYTHON_INCLUDE_CLI = None,
    python_version: PYTHON_VERSION_CLI = None,
    python: PYTHON_CLI = None,
    skip_package: SKIP_PACKAGE_CLI = False,
    sort: SORT_DEPENDENCIES_CLI = True,
    header: HEADER_CLI = None,
    overwrite: OVERWRITE_CLI = Overwrite.check,
    verbose: VERBOSE_CLI = None,
    deps: DEPS_CLI = None,
    reqs: REQS_CLI = None,
    allow_empty: Annotated[bool, ALLOW_EMPTY_OPTION] = False,
    remove_whitespace: Annotated[bool, REMOVE_WHITESPACE_OPTION] = True,
) -> None:
    """Create yaml file from dependencies and optional-dependencies."""
    if not update_target(output, filename, overwrite=overwrite.value):
        _log_skipping(logger, "yaml", output)
        return

    if not channels:
        channels = None

    python_include, python_version = parse_pythons(
        python_include=python_include,
        python_version=python_version,
        python=python,
        toml_path=filename,
    )

    d = _get_requirement_parser(filename)

    _log_creating(logger, "yaml", output)

    s = d.to_conda_yaml(
        extras=extras,
        groups=groups,
        extras_or_groups=extras_or_groups,
        channels=channels,
        name=name,
        output=output,
        python_include=python_include,
        python_version=python_version,
        skip_package=skip_package,
        header_cmd=_get_header_cmd(header, output),
        sort=sort,
        conda_deps=deps,
        pip_deps=reqs,
        allow_empty=allow_empty,
        remove_whitespace=remove_whitespace,
    )
    if not output:
        print(s, end="")


# ** Requirements
# @app_typer.command("r", hidden=True)
@app_typer.command()
@add_verbose_logger(logger)
def requirements(
    filename: PYPROJECT_CLI = DEFAULT_TOML_PATH,
    extras: EXTRAS_CLI = None,
    groups: GROUPS_CLI = None,
    extras_or_groups: EXTRAS_OR_GROUPS_CLI = None,
    output: OUTPUT_CLI = None,
    skip_package: SKIP_PACKAGE_CLI = False,
    sort: SORT_DEPENDENCIES_CLI = True,
    header: HEADER_CLI = None,
    overwrite: OVERWRITE_CLI = Overwrite.check,
    verbose: VERBOSE_CLI = None,
    reqs: REQS_CLI = None,
    allow_empty: Annotated[bool, ALLOW_EMPTY_OPTION] = False,
    remove_whitespace: Annotated[bool, REMOVE_WHITESPACE_OPTION] = False,
) -> None:
    """Create requirements.txt for pip dependencies.  Note that all requirements are normalized using `packaging.requirements.Requirement`"""
    if not update_target(output, filename, overwrite=overwrite.value):
        _log_skipping(logger, "requirements", output)
        return

    d = _get_requirement_parser(filename)

    _log_creating(logger, "requirements", output)

    s = d.to_requirements(
        extras=extras,
        groups=groups,
        extras_or_groups=extras_or_groups,
        output=output,
        skip_package=skip_package,
        header_cmd=_get_header_cmd(header, output),
        sort=sort,
        pip_deps=reqs,
        allow_empty=allow_empty,
        remove_whitespace=remove_whitespace,
    )
    if not output:
        print(s, end="")


# ** From project


# @app_typer.command("p", hidden=True)
@app_typer.command()
@add_verbose_logger(logger)
def project(
    filename: PYPROJECT_CLI = DEFAULT_TOML_PATH,
    envs: ENVS_CLI = None,
    template: TEMPLATE_CLI = None,
    template_python: TEMPLATE_PYTHON_CLI = None,
    reqs: REQS_CLI = None,
    deps: DEPS_CLI = None,
    reqs_ext: REQS_EXT_CLI = ".txt",
    yaml_ext: YAML_EXT_CLI = ".yaml",
    sort: SORT_DEPENDENCIES_CLI = True,
    header: HEADER_CLI = None,
    overwrite: OVERWRITE_CLI = Overwrite.check,
    verbose: VERBOSE_CLI = None,
    dry: DRY_CLI = False,
    user_config: USER_CONFIG_CLI = "infer",
    allow_empty: Annotated[Optional[bool], ALLOW_EMPTY_OPTION] = None,
    remove_whitespace: Annotated[Optional[bool], REMOVE_WHITESPACE_OPTION] = None,
) -> None:
    """Create multiple environment files from `pyproject.toml` specification."""
    from pyproject2conda.config import Config

    c = Config.from_file(filename, user_config=user_config)

    if user_config == "infer" or user_config is None:
        user_config = c.user_config()

    for style, d in c.iter_envs(
        envs=envs,
        reqs_ext=reqs_ext,
        yaml_ext=yaml_ext,
        template=template,
        template_python=template_python,
        reqs=reqs,
        deps=deps,
        sort=sort,
        header=header,
        overwrite=overwrite.value,
        verbose=verbose,
        allow_empty=allow_empty,
        remove_whitespace=remove_whitespace,
    ):
        if dry:
            # small header
            print("# " + "-" * 20)
            print("# Creating {style} {output}".format(style=style, output=d["output"]))
            d["output"] = None

        # Special case: have output and userconfig.  Check update
        if not update_target(
            d["output"],
            filename,
            *([user_config] if user_config else []),
            overwrite=d["overwrite"],
        ):
            if verbose:
                _log_skipping(logger, style, d["output"])
        else:
            d["overwrite"] = Overwrite("force")
            if style == "yaml":
                yaml(filename=filename, **d)

            elif style == "requirements":
                requirements(filename=filename, **d)
            else:  # pragma: no cover
                msg = f"unknown style {style}"
                raise ValueError(msg)


# ** Conda requirements


# @app_typer.command("cr", hidden=True)
@app_typer.command()
@add_verbose_logger(logger)
def conda_requirements(
    path_conda: Annotated[Optional[str], typer.Argument()] = None,
    path_pip: Annotated[Optional[str], typer.Argument()] = None,
    extras: EXTRAS_CLI = None,
    groups: GROUPS_CLI = None,
    extras_or_groups: EXTRAS_OR_GROUPS_CLI = None,
    python_include: PYTHON_INCLUDE_CLI = None,
    python_version: PYTHON_VERSION_CLI = None,
    python: PYTHON_CLI = None,
    channels: CHANNEL_CLI = None,
    filename: PYPROJECT_CLI = DEFAULT_TOML_PATH,
    skip_package: SKIP_PACKAGE_CLI = False,
    prefix: PREFIX_CLI = None,
    prepend_channel: PREPEND_CHANNEL_CLI = False,
    sort: SORT_DEPENDENCIES_CLI = True,
    header: HEADER_CLI = None,
    # paths,
    deps: DEPS_CLI = None,
    reqs: REQS_CLI = None,
    verbose: VERBOSE_CLI = None,
) -> None:
    """
    Create requirement files for conda and pip.

    These can be install with, for example,

    conda install --file {path_conda}
    pip install -r {path_pip}
    """
    python_include, python_version = parse_pythons(
        python_include=python_include,
        python_version=python_version,
        python=python,
        toml_path=filename,
    )

    if path_conda and not path_pip:
        msg = "can only specify neither or both path_conda and path_pip"
        raise ValueError(msg)

    if path_conda and path_pip and prefix is not None:
        msg = "specify path_conda and path_pip or prefix, not all"
        raise ValueError(msg)

    if prefix is not None:
        path_conda = prefix + "conda.txt"
        path_pip = prefix + "pip.txt"

    d = _get_requirement_parser(filename)

    _get_header_cmd(header, path_conda)

    deps_str, reqs_str = d.to_conda_requirements(
        extras=extras,
        groups=groups,
        extras_or_groups=extras_or_groups,
        python_include=python_include,
        python_version=python_version,
        channels=channels,
        prepend_channel=prepend_channel,
        output_conda=path_conda,
        output_pip=path_pip,
        skip_package=skip_package,
        header_cmd=_get_header_cmd(header, path_conda),
        sort=sort,
        conda_deps=deps,
        pip_deps=reqs,
    )

    if not path_conda:
        s = f"#conda requirements\n{deps_str}\n#pip requirements\n{reqs_str}"
        print(s, end="")


# ** json
# @app_typer.command("j", hidden=True)
@app_typer.command("json")
@add_verbose_logger(logger)
def to_json(
    extras: EXTRAS_CLI = None,
    groups: GROUPS_CLI = None,
    extras_or_groups: EXTRAS_OR_GROUPS_CLI = None,
    python_include: PYTHON_INCLUDE_CLI = None,
    python_version: PYTHON_VERSION_CLI = None,
    python: PYTHON_CLI = None,
    channels: CHANNEL_CLI = None,
    filename: PYPROJECT_CLI = DEFAULT_TOML_PATH,
    sort: SORT_DEPENDENCIES_CLI = True,
    output: OUTPUT_CLI = None,
    skip_package: SKIP_PACKAGE_CLI = False,
    deps: DEPS_CLI = None,
    reqs: REQS_CLI = None,
    verbose: VERBOSE_CLI = None,
    overwrite: OVERWRITE_CLI = Overwrite.check,
) -> None:
    """
    Create json representation.

    Keys are:
    "dependencies": conda dependencies.
    "pip": pip dependencies.
    "channels": conda channels.
    """
    if not update_target(output, filename, overwrite=overwrite.value):
        _log_skipping(logger, "yaml", output)
        return

    import json

    d = _get_requirement_parser(filename)

    python_include, python_version = parse_pythons(
        python_include=python_include,
        python_version=python_version,
        python=python,
        toml_path=filename,
    )

    conda_deps, pip_deps = d.conda_and_pip_requirements(
        extras=extras,
        groups=groups,
        extras_or_groups=extras_or_groups,
        python_include=python_include,
        python_version=python_version,
        skip_package=skip_package,
        sort=sort,
        conda_deps=deps,
        pip_deps=reqs,
    )

    result = {
        "dependencies": conda_deps,
        "pip": pip_deps,
    }

    channels = channels or d.channels
    if channels:
        result["channels"] = channels

    if output:
        with Path(output).open("w", encoding=locale.getpreferredencoding(False)) as f:
            json.dump(result, f)
    else:
        print(json.dumps(result))  # , indent=2))


# * Click app
# # If need be, can work directly with click
# @click.group(cls=AliasedGroup)
# @click.version_option(version=__version__)
# def app_click() -> None:
#     pass
# typer_click_object = typer.main.get_command(app_typer)  # noqa: ERA001
# app = click.CommandCollection(sources=[app_click, typer_click_object], cls=AliasedGroup)  # noqa: ERA001

# Just use the click app....
app = typer.main.get_command(app_typer)


# ** Main
if __name__ == "__main__":
    app()  # pragma: no cover



"""Config file for nox."""

# * Imports ----------------------------------------------------------------------------
from __future__ import annotations

import shlex
import shutil
import sys
from functools import lru_cache, partial, wraps

# Should only use on python version > 3.10
if sys.version_info < (3, 10):
    msg = "python>=3.10 required"
    raise RuntimeError(msg)

from dataclasses import dataclass
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Literal,
    TypeAlias,
    TypedDict,
)

# fmt: off
sys.path.insert(0, ".")
from tools import uvxrun
from tools.dataclass_parser import DataclassParser, add_option, option
from tools.noxtools import (
    Installer,
    add_uv_pythons_to_path,
    cached_which,
    check_for_change_manager,
    combine_list_list_str,
    combine_list_str,
    factory_conda_backend,
    factory_virtualenv_backend,
    infer_requirement_path,
    is_conda_session,
    open_webpage,
    prepend_flag,
    session_run_commands,
    update_target,
)

sys.path.pop(0)

# make sure these afeter
import nox  # type: ignore[unused-ignore,import]

# fmt: on

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Sequence

    from nox import Session
    from nox.virtualenv import CondaEnv


# * Names ------------------------------------------------------------------------------

PACKAGE_NAME = "pyproject2conda"
IMPORT_NAME = "pyproject2conda"
KERNEL_BASE = "pyproject2conda"

# * nox options ------------------------------------------------------------------------

ROOT = Path(__file__).parent

nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = ["test"]
nox.options.default_venv_backend = "uv"

# * User Config ------------------------------------------------------------------------

# CONFIG = load_nox_config()  # noqa: ERA001
# if you'd like to disallow uv.
# You'll need to import this from tools.noxtools
# DISALLOW_WHICH.append("uv")  # noqa: ERA001

add_uv_pythons_to_path()


# * Options ---------------------------------------------------------------------------

LOCK = True

PYTHON_ALL_VERSIONS = ["3.8", "3.9", "3.10", "3.11", "3.12", "3.13"]
PYTHON_DEFAULT_VERSION = "3.11"


UVXRUN_LOCK_REQUIREMENTS = "requirements/lock/py{}-uvxrun-tools.txt".format(
    PYTHON_DEFAULT_VERSION.replace(".", "")
)
UVXRUN_MIN_REQUIREMENTS = "requirements/uvxrun-tools.txt"


@lru_cache
def get_uvxrun_specs(requirements: str | None = None) -> uvxrun.Specifications:
    """Get specs for uvxrun."""
    requirements = requirements or UVXRUN_MIN_REQUIREMENTS
    if not Path(requirements).exists():
        requirements = None
    return uvxrun.Specifications.from_requirements(requirements=requirements)


for backend in ["mamba", "micromamba", "conda"]:
    if shutil.which(backend):
        CONDA_BACKEND: Literal["mamba", "micromamba", "conda"] = backend  # type: ignore[assignment]
        break
else:
    msg = "no conda-like backend found"
    raise ValueError(msg)


class SessionOptionsDict(TypedDict, total=False):
    """Dict for options to nox.session"""

    python: str | list[str]
    venv_backend: str | Callable[..., CondaEnv]


CONDA_DEFAULT_KWS: SessionOptionsDict = {
    "python": PYTHON_DEFAULT_VERSION,
    "venv_backend": factory_conda_backend(CONDA_BACKEND),
}
CONDA_ALL_KWS: SessionOptionsDict = {
    "python": PYTHON_ALL_VERSIONS,
    "venv_backend": factory_conda_backend(CONDA_BACKEND),
}

DEFAULT_KWS: SessionOptionsDict = {"python": PYTHON_DEFAULT_VERSION}
ALL_KWS: SessionOptionsDict = {"python": PYTHON_ALL_VERSIONS}


# * Session command line options -------------------------------------------------------

OPT_TYPE: TypeAlias = list[str] | None
RUN_TYPE: TypeAlias = list[list[str]] | None

RUN_ANNO = Annotated[
    RUN_TYPE,
    option(
        help="Run external commands in session.  Specify multiple times for multiple commands.",
    ),
]
OPT_ANNO = Annotated[OPT_TYPE, option(help="Options to command.")]


@dataclass
class SessionParams(DataclassParser):
    """Holds all cli options"""

    # common parameters
    lock: bool = False
    update: bool = add_option("--update", "-U", help="update dependencies/package")
    update_package: bool = add_option(
        "--update-package",
        "-P",
        help="update package only",
    )
    log_session: bool = add_option("--log-session")
    version: str | None = None

    # dev
    dev_run: RUN_ANNO = None
    dev_envname: Literal["dev", "dev-complete", "dev-user"] = add_option(
        help="Name of environment to use for development session",
        default="dev",
    )

    # config
    dev_extras: OPT_TYPE = add_option(help="`extras` to include in dev environment")
    python_paths: OPT_TYPE = add_option(help="paths to python executables")

    # requirements
    requirements_force: bool = False
    requirements_no_notify: bool = add_option(
        default=False,
        help="Skip notification of pip-compile",
    )

    # conda-lock
    conda_lock_channel: OPT_TYPE = add_option(help="conda channels")
    conda_lock_platform: (
        list[Literal["osx-64", "linux-64", "win-64", "osx-arm64", "all"]] | None
    ) = add_option(help="platform(s) to buiuld lock file for.")
    conda_lock_include: OPT_TYPE = add_option(help="lock files to create")
    conda_lock_mamba: bool = False
    conda_lock_force: bool = False

    # pip-compile
    pip_compile_force: bool = False
    pip_compile_upgrade: bool = add_option(
        "--pip-compile-upgrade",
        "-L",
        help="Upgrade all packages in lock file",
        default=False,
    )
    pip_compile_upgrade_package: OPT_TYPE = add_option(
        help="Upgrade package(s) in lock file",
        default=None,
    )
    pip_compile_opts: OPT_TYPE = add_option(help="options to pip-compile")
    pip_compile_run: RUN_ANNO = None
    pip_compile_run_internal: RUN_ANNO = None

    # test
    test_no_pytest: bool = False
    test_opts: OPT_TYPE = add_option(help="Options to pytest")
    test_run: RUN_ANNO = None
    no_cov: bool = False

    # coverage
    coverage: list[Literal["erase", "combine", "report", "html", "open"]] | None = None

    # testdist
    testdist_run: RUN_ANNO = None

    # docs
    docs: (
        list[
            Literal[
                "html",
                "build",
                "symlink",
                "clean",
                "livehtml",
                "linkcheck",
                "spelling",
                "showlinks",
                "open",
                "serve",
            ]
        ]
        | None
    ) = add_option("--docs", "-d", help="doc commands")
    docs_run: RUN_ANNO = None

    # typing
    typing: list[
        Literal[
            "clean",
            "mypy",
            "pyright",
            "pytype",
            "all",
            "mypy-notebook",
            "pyright-notebook",
            "typecheck-notebook",
        ]
    ] = add_option("--typing", "-m")
    typing_run: RUN_ANNO = None
    typing_run_internal: RUN_TYPE = add_option(
        help="Run internal (in session) commands.",
    )

    # build
    build: list[Literal["build", "version"]] | None = None
    build_run: RUN_ANNO = None
    build_isolation: bool = False
    build_outdir: str = "./dist"
    build_opts: OPT_ANNO = None
    build_silent: bool = False

    # publish
    publish: list[Literal["release", "test", "check"]] | None = add_option(
        "-p", "--publish"
    )

    # conda-recipe/grayskull
    conda_recipe: list[Literal["recipe", "recipe-full"]] | None = None
    conda_recipe_sdist_path: str | None = None

    # conda-build
    conda_build: list[Literal["build", "clean"]] | None = None
    conda_build_run: RUN_ANNO = None


@lru_cache
def parse_posargs(*posargs: str) -> SessionParams:
    """
    Get Parser using `+` for session option prefix.

    Note that using `+` allows for passing underlying `-` options
    without escaping.
    """
    opts = SessionParams.from_posargs(posargs=posargs, prefix_char="+")
    opts.lock = opts.lock or LOCK

    return opts


def add_opts(
    func: Callable[[Session, SessionParams], None],
) -> Callable[[Session], None]:
    """Fill in `opts` from cli options."""

    @wraps(func)
    def wrapped(session: Session) -> None:
        opts = parse_posargs(*session.posargs)
        return func(session, opts)

    return wrapped


# * Environments------------------------------------------------------------------------
# ** Dev (conda)
@add_opts
def dev(
    session: Session,
    opts: SessionParams,
) -> None:
    """Create development environment using either conda (dev) or virtualenv (dev-venv) in location `.venv`"""
    (
        Installer.from_envname(
            session=session,
            envname=opts.dev_envname,
            lock=opts.lock,
            update=opts.update,
            package=True,
        )
        .install_all(
            update_package=opts.update_package,
            log_session=opts.log_session,
        )
        .set_ipykernel_display_name(name=f"{PACKAGE_NAME}-dev", update=True)
        .run_commands(opts.dev_run)
    )


nox.session(
    name="dev",
    python=PYTHON_DEFAULT_VERSION,
    venv_backend=factory_virtualenv_backend(location="./.venv"),
)(dev)
nox.session(
    name="dev-conda",
    python=PYTHON_DEFAULT_VERSION,
    venv_backend=factory_conda_backend(backend=CONDA_BACKEND, location="./.venv"),
)(dev)


@nox.session(python=False)
@add_opts
def config(
    session: Session,
    opts: SessionParams,
) -> None:
    """Create the file ./config/userconfig.toml"""
    args: list[str] = []
    if opts.dev_extras:
        args += ["--dev-extras", *opts.dev_extras]
    if opts.python_paths:
        args += ["--python-paths", *opts.python_paths]

    session.run(sys.executable, "tools/projectconfig.py", *args)


# ** requirements
@nox.session(name="requirements", python=False)
@add_opts
def requirements(
    session: Session,
    opts: SessionParams,
) -> None:
    """
    Create environment.yaml and requirement.txt files from pyproject.toml using pyproject2conda.

    These will be placed in the directory "./requirements".
    """
    uvxrun.run(
        "pyproject2conda",
        "project",
        "--verbose",
        *(["--overwrite=force"] if opts.requirements_force else []),
        session=session,
        external=True,
        specs=get_uvxrun_specs(),
    )

    if not opts.requirements_no_notify and opts.lock:
        if cached_which("uv"):
            session.notify("uv-compile")
        else:
            for py in PYTHON_ALL_VERSIONS:
                session.notify(f"pip-compile-{py}")


# # ** conda-lock
@nox.session(name="conda-lock", python=False)
@add_opts
def conda_lock(
    session: Session,
    opts: SessionParams,
) -> None:
    """Create lock files using conda-lock."""
    conda_lock_exclude = ["test-extras"]
    conda_lock_include = opts.conda_lock_include or [
        "test",
        "dev",
        "dev-complete",
        "nox",
    ]

    platform = opts.conda_lock_platform
    if platform is None or "all" in platform:
        # for now, skip osx-arm64 and win-64.  Leads to some problems.
        platform = ["osx-64", "linux-64"]  # , "win-64"]

    channel = opts.conda_lock_channel
    if not channel:
        channel = ["conda-forge"]

    def create_lock(path: Path) -> None:
        name = path.with_suffix("").name
        lockfile = path.parent / "lock" / f"{name}-conda-lock.yml"
        deps = [str(path)]

        # check if skip
        env = "-".join(name.split("-")[1:])
        if conda_lock_include and not any(c == env for c in conda_lock_include):
            session.log(f"Skipping {lockfile} (include)")
            return

        if conda_lock_exclude and any(c == env for c in conda_lock_exclude):
            session.log(f"Skipping {lockfile} (exclude)")
            return

        # check hashes

        with check_for_change_manager(
            *deps,
            target_path=lockfile,
            force_write=opts.conda_lock_force,
        ) as changed:
            if opts.conda_lock_force or changed:
                session.log(f"Creating {lockfile}")
                # insert -f for each arg
                if lockfile.exists():
                    lockfile.unlink()
                uvxrun.run(
                    "conda-lock",
                    "--mamba" if opts.conda_lock_mamba else "--no-mamba",
                    *prepend_flag("-c", *channel),
                    *prepend_flag("-p", *platform),
                    *prepend_flag("-f", *deps),
                    f"--lockfile={lockfile}",
                    specs=get_uvxrun_specs(),
                    session=session,
                    external=True,
                )
            else:
                session.log(f"Skipping {lockfile} (exists)")

    for path in (ROOT / "requirements").relative_to(ROOT.cwd()).glob("py*.yaml"):
        create_lock(path)


def _run_compile_pre(
    runner: Installer,
    run: RUN_TYPE,
    run_internal: RUN_TYPE,
) -> bool:
    if run:
        runner.run_commands(run)
        return True

    if run_internal:
        runner.run_commands(run_internal, external=False)
        return True

    return False


def _run_compile_options(
    options: OPT_TYPE,
    force: bool,
    upgrade: bool,
    upgrade_package: OPT_TYPE,
) -> tuple[list[str], bool]:
    options = list(options) if options else []
    if upgrade:
        options = [*options, "-U"]

    if upgrade_package:
        options = [*options, *prepend_flag("-P", upgrade_package)]

    force = force or upgrade or bool(upgrade_package)

    return options, force


def _run_compile_env(
    session: nox.Session,
    compile_command: Sequence[str],
    python: str,
    options: Sequence[str],
    force: bool,
    env: str,
    envs_optional: Sequence[str],
    **kwargs: Any,
) -> None:
    reqspath = infer_requirement_path(env, ext=".txt", check_exists=False)
    if not reqspath.is_file():
        if env in envs_optional:
            return
        msg = f"Missing file {reqspath}"
        raise ValueError(msg)

    lockpath = infer_requirement_path(
        env,
        ext=".txt",
        python_version=python,
        lock=True,
        check_exists=False,
    )

    with check_for_change_manager(
        reqspath,
        target_path=lockpath,
        force_write=force,
    ) as changed:
        if force or changed:
            session.log(f"Creating {lockpath}")
            session.run(
                *compile_command, *options, "-o", str(lockpath), str(reqspath), **kwargs
            )

        else:
            session.log(f"Skipping {lockpath}")


@nox.session(name="pip-compile", **ALL_KWS)
@add_opts
def pip_compile(
    session: Session,
    opts: SessionParams,
) -> None:
    """
    Run pip-compile.

    Note that this session is also used to run pip-sync with correct python version for
    tests/typing/etc.
    """
    runner = Installer(
        session=session,
        pip_deps=["pip-tools"],
        update=opts.update,
    ).install_all(log_session=opts.log_session)

    if not isinstance(session.python, str):
        msg = "must set python version"
        raise TypeError(msg)

    if _run_compile_pre(runner, opts.pip_compile_run, opts.pip_compile_run_internal):
        return

    options, force = _run_compile_options(
        options=opts.pip_compile_opts,
        force=opts.pip_compile_force,
        upgrade=opts.pip_compile_upgrade,
        upgrade_package=opts.pip_compile_upgrade_package,
    )

    envs_all = ["test", "typing"]
    envs_dev = ["dev", "dev-complete", "docs"]
    envs_dev_optional = ["test-notebook", "uvxrun-tools"]

    if session.python == PYTHON_DEFAULT_VERSION:
        envs = envs_all + envs_dev + envs_dev_optional
    else:
        envs = envs_all

    for env in envs:
        _run_compile_env(
            session=session,
            compile_command=("pip-compile",),
            python=session.python,
            options=options,
            force=force,
            env=env,
            envs_optional=envs_dev_optional,
            external=False,
        )


# ** uv pip compile
@nox.session(name="uv-compile", python=False)
@add_opts
def uv_compile(
    session: Session,
    opts: SessionParams,
) -> None:
    """Run uv pip compile ..."""
    uv_path = cached_which("uv")
    if uv_path is None:
        session.log("Need to install uv to use it...")
        return

    options, force = _run_compile_options(
        options=opts.pip_compile_opts,
        force=opts.pip_compile_force,
        upgrade=opts.pip_compile_upgrade,
        upgrade_package=opts.pip_compile_upgrade_package,
    )

    envs_all = ["test", "typing"]
    envs_dev = ["dev", "dev-complete", "docs"]
    envs_dev_optional = ["test-notebook", "uvxrun-tools"]

    for python in set(PYTHON_ALL_VERSIONS).union({PYTHON_DEFAULT_VERSION}):
        if python == PYTHON_DEFAULT_VERSION:
            envs = envs_all + envs_dev + envs_dev_optional
        else:
            envs = envs_all

        for env in envs:
            _run_compile_env(
                session=session,
                compile_command=(
                    uv_path,
                    "pip",
                    "compile",
                    f"--python-version={python}",
                ),
                python=python,
                options=options,
                force=force,
                env=env,
                envs_optional=envs_dev_optional,
                external=True,
            )


# ** testing
def _test(
    session: nox.Session,
    run: RUN_TYPE,
    test_no_pytest: bool,
    test_opts: OPT_TYPE,
    no_cov: bool,
) -> None:
    import os

    tmpdir = os.environ.get("TMPDIR", None)

    session_run_commands(session, run)
    if not test_no_pytest:
        opts = combine_list_str(test_opts or [])
        if not no_cov:
            session.env["COVERAGE_FILE"] = str(Path(session.create_tmp()) / ".coverage")

            if not any(o.startswith("--cov") for o in opts):
                opts.append(f"--cov={IMPORT_NAME}")

        # Because we are testing if temporary folders
        # have git or not, we have to make sure we're above the
        # not under this repo
        # so revert to using the top level `TMPDIR`
        if tmpdir:
            session.env["TMPDIR"] = tmpdir

        session.run("pytest", *opts)


# *** Basic tests
@add_opts
def test(
    session: Session,
    opts: SessionParams,
) -> None:
    """Test environments with conda installs."""
    (
        Installer.from_envname(
            session=session,
            envname="test",
            lock=opts.lock,
            # To use editable install
            # package=True,  # noqa: ERA001
            # To use full install
            package=get_package_wheel(session, opts="--no-deps --force-reinstall"),
            update=opts.update,
        ).install_all(log_session=opts.log_session, update_package=opts.update_package)
    )

    _test(
        session=session,
        run=opts.test_run,
        test_no_pytest=opts.test_no_pytest,
        test_opts=opts.test_opts,
        no_cov=opts.no_cov,
    )


nox.session(name="test", **ALL_KWS)(test)
nox.session(name="test-conda", **CONDA_ALL_KWS)(test)


@nox.session(name="test-notebook", **DEFAULT_KWS)
@add_opts
def test_notebook(session: nox.Session, opts: SessionParams) -> None:
    """Run pytest --nbval."""
    (
        Installer.from_envname(
            session=session,
            envname="test-notebook",
            lock=opts.lock,
            package=get_package_wheel(session, opts="--no-deps --force-reinstall"),
            update=opts.update,
        ).install_all(log_session=opts.log_session, update_package=opts.update_package)
    )

    test_nbval_opts = shlex.split(
        """
    --nbval
    --nbval-current-env
    --nbval-sanitize-with=config/nbval.ini
    --dist loadscope
     examples/usage
   """,
    )

    test_opts = (opts.test_opts or []) + test_nbval_opts

    session.log(f"{test_opts = }")

    _test(
        session=session,
        run=opts.test_run,
        test_no_pytest=opts.test_no_pytest,
        test_opts=test_opts,
        no_cov=opts.no_cov,
    )


@nox.session(python=False)
@add_opts
def coverage(
    session: Session,
    opts: SessionParams,
) -> None:
    """Run coverage."""
    cmd = opts.coverage or ["combine", "html", "report"]

    run = partial(uvxrun.run, specs=get_uvxrun_specs(), session=session)

    paths = list(Path(".nox").glob("test-*/tmp/.coverage*"))

    if "erase" in cmd:
        for path in paths:
            if path.exists():
                session.log(f"removing {path}")
                path.unlink()

    for c in cmd:
        if c == "combine":
            if update_target(".coverage", *paths):
                run(
                    "coverage",
                    "combine",
                    "--keep",
                    "-a",
                    *paths,
                )
        elif c == "open":
            open_webpage(path="htmlcov/index.html")

        else:
            run(
                "coverage",
                c,
            )


# *** testdist (conda)
def testdist(
    session: Session,
) -> None:
    """Test conda distribution."""
    opts = parse_posargs(*session.posargs)

    install_str = PACKAGE_NAME
    if opts.version:
        install_str = f"{install_str}=={opts.version}"

    if is_conda_session(session):
        pip_deps, conda_deps = None, install_str
    else:
        pip_deps, conda_deps = install_str, None

    (
        Installer.from_envname(
            session=session,
            envname="test-extras",
            conda_deps=conda_deps,
            pip_deps=pip_deps,
            update=opts.update,
            channels="conda-forge",
        ).install_all(log_session=opts.log_session)
    )

    _test(
        session=session,
        run=opts.testdist_run,
        test_no_pytest=opts.test_no_pytest,
        test_opts=opts.test_opts,
        no_cov=opts.no_cov,
    )


nox.session(name="testdist-pypi", **ALL_KWS)(testdist)
nox.session(name="testdist-conda", **CONDA_ALL_KWS)(testdist)


# # ** Docs
@add_opts
def docs(
    session: nox.Session,
    opts: SessionParams,
) -> None:
    """
    Run `make` in docs directory.

    For example, 'nox -s docs -- +d html'
    calls 'make -C docs html'. With 'release' option, you can set the
    message with 'message=...' in posargs.
    """
    runner = Installer.from_envname(
        session=session,
        envname="docs",
        lock=opts.lock,
        package=True,
        update=opts.update,
    ).install_all(
        update_package=opts.update_package,
        log_session=opts.log_session,
    )

    if opts.version:
        session.env["SETUPTOOLS_SCM_PRETEND_VERSION"] = opts.version

    runner.run_commands(opts.docs_run)

    cmd = opts.docs or []
    cmd = ["html"] if not opts.docs_run and not cmd else list(cmd)

    if "symlink" in cmd:
        cmd.remove("symlink")
        _create_doc_examples_symlinks(session)

    if open_page := "open" in cmd:
        cmd.remove("open")

    if serve := "serve" in cmd:
        open_webpage(url="http://localhost:8000")
        cmd.remove("serve")

    if cmd:
        args = ["make", "-C", "docs", *combine_list_str(cmd)]
        session.run(*args, external=True)

    if open_page:
        open_webpage(path="./docs/_build/html/index.html")

    if serve and "livehtml" not in cmd:
        session.run(
            "python",
            "-m",
            "http.server",
            "-d",
            "docs/_build/html",
            "-b",
            "127.0.0.1",
            "8000",
        )


nox.session(name="docs", **DEFAULT_KWS)(docs)
nox.session(name="docs-conda", **CONDA_DEFAULT_KWS)(docs)


# ** lint
@nox.session(python=False)
def lint(
    session: nox.Session,
) -> None:
    """
    Run linters with pre-commit.

    Defaults to `pre-commit run --all-files`.
    To run something else pass, e.g.,
    `nox -s lint -- --lint-run "pre-commit run --hook-stage manual --all-files`
    """
    uvxrun.run(
        "pre-commit",
        "run",
        "--all-files",  # "--show-diff-on-failure",
        specs=get_uvxrun_specs(),
        session=session,
    )


# ** type checking
@add_opts
def typing(  # noqa: C901
    session: nox.Session,
    opts: SessionParams,
) -> None:
    """Run type checkers (mypy, pyright, pytype)."""
    runner = (
        Installer.from_envname(
            session=session,
            envname="typing",
            lock=opts.lock,
            update=opts.update,
            # need package for nbqa checks
            package=True,
        )
        .install_all(log_session=opts.log_session)
        .run_commands(opts.typing_run)
    )

    cmd = opts.typing or []
    if not opts.typing_run and not opts.typing_run_internal and not cmd:
        cmd = ["mypy", "pyright"]

    if "all" in cmd:
        cmd = ["mypy", "pyright", "pytype"]

    # set the cache directory for mypy
    session.env["MYPY_CACHE_DIR"] = str(Path(session.create_tmp()) / ".mypy_cache")

    if "clean" in cmd:
        cmd = list(cmd)
        cmd.remove("clean")

        for name in [".mypy_cache", ".pytype"]:
            p = Path(session.create_tmp()) / name
            if p.exists():
                session.log(f"removing cache {p}")
                shutil.rmtree(str(p))

    if not isinstance(session.python, str):
        raise TypeError

    run = partial(
        uvxrun.run,
        specs=get_uvxrun_specs(UVXRUN_LOCK_REQUIREMENTS),
        session=session,
        python_version=session.python,
        python_executable=runner.python_full_path,
        external=True,
    )

    for c in cmd:
        if c.endswith("-notebook"):
            session.run("make", c, external=True)
        elif c == "mypy":
            run("mypy", "--color-output")
        elif c == "pyright":
            run("pyright")
        else:
            session.log(f"Skipping unknown command {c}")

    for cmds in combine_list_list_str(opts.typing_run_internal or []):
        run(*cmds)


nox.session(name="typing", **ALL_KWS)(typing)
nox.session(name="typing-conda", **CONDA_ALL_KWS)(typing)


# ** Dist pypi
# NOTE: you can skip having the build environment and
# just use uv build, but faster to use environment ...
USE_ENVIRONMENT_FOR_BUILD = False
_build_dec = nox.session if USE_ENVIRONMENT_FOR_BUILD else nox.session(python=False)


@_build_dec
@add_opts
def build(session: nox.Session, opts: SessionParams) -> None:  # noqa: C901
    """
    Build the distribution.

    Note that default is to not use build isolation.
    Pass `--build-isolation` to use build isolation.
    """
    if USE_ENVIRONMENT_FOR_BUILD:
        runner = Installer.from_envname(
            session=session,
            envname="build",
            update=opts.update,
        ).install_all(log_session=opts.log_session)

    if opts.version:
        session.env["SETUPTOOLS_SCM_PRETEND_VERSION"] = opts.version

    for cmd in opts.build or ["build"]:
        if cmd == "version":
            if USE_ENVIRONMENT_FOR_BUILD:
                session.run(runner.python_full_path, "-m", "hatchling", "version")  # pyright: ignore[reportPossiblyUnboundVariable]
            else:
                session.run(
                    "uvx", "--with", "hatch-vcs", "hatchling", "version", external=True
                )

        elif cmd == "build":
            outdir = opts.build_outdir
            if Path(outdir).exists():
                shutil.rmtree(outdir)

            args = f"uv build --out-dir {outdir}".split()
            if USE_ENVIRONMENT_FOR_BUILD and not opts.build_isolation:
                args.append("--no-build-isolation")

            if opts.build_opts:
                args.extend(opts.build_opts)

            out = session.run(*args, silent=opts.build_silent)
            if opts.build_silent:
                if not isinstance(out, str):
                    msg = "session.run output not a string"
                    raise ValueError(msg)
                session.log(out.strip().split("\n")[-1])


def get_package_wheel(
    session: Session,
    opts: str | Iterable[str] | None = None,
    extras: str | Iterable[str] | None = None,
    reuse: bool = True,
) -> str:
    """
    Build the package in return the build location

    This is similar to how tox does isolated builds.

    Note that the first time this is called,

    Should be straightforward to extend this to isolated builds
    that depend on python version (something like have session build-3.11 ....)
    """
    dist_location = Path(session.cache_dir) / "dist"
    if reuse and getattr(get_package_wheel, "_called", False):
        session.log("Reuse isolated build")
    else:
        cmd = f"nox -s build -- ++build-outdir {dist_location} ++build-opts --wheel ++build-silent"
        session.run_always(*shlex.split(cmd), external=True)

        # save that this was called:
        if reuse:
            get_package_wheel._called = True  # type: ignore[attr-defined]  # noqa: SLF001

    paths = list(dist_location.glob("*.whl"))
    if len(paths) != 1:
        msg = f"something wonky with paths {paths}"
        raise ValueError(msg)

    path = str(paths[0])

    if cached_which("uv"):
        path = f"{PACKAGE_NAME}@{path}"

    if extras:
        if not isinstance(extras, str):
            extras = ",".join(extras)
        path = f"{path}[{extras}]"

    if opts:
        if not isinstance(opts, str):
            opts = " ".join(opts)
        path = f"{path} {opts}"

    return path


@nox.session(python=False)
@add_opts
def publish(session: nox.Session, opts: SessionParams) -> None:
    """Publish the distribution"""
    run = partial(uvxrun.run, specs=get_uvxrun_specs(), session=session, external=True)

    for cmd in opts.publish or []:
        if cmd == "test":
            run("twine", "upload", "--repository", "testpypi", "dist/*")
        elif cmd == "release":
            run("twine", "upload", "dist/*")
        elif cmd == "check":
            run("twine", "check", "--strict", "dist/*")


# # ** Dist conda
@nox.session(name="conda-recipe", python=False)
@add_opts
def conda_recipe(
    session: nox.Session,
    opts: SessionParams,
) -> None:
    """Run grayskull to create recipe"""
    commands = opts.conda_recipe or ["recipe"]

    run = partial(uvxrun.run, specs=get_uvxrun_specs(), session=session)

    if not (sdist_path := opts.conda_recipe_sdist_path):
        sdist_path = PACKAGE_NAME
        if opts.version:
            sdist_path = f"{sdist_path}=={opts.version}"

    for command in commands:
        if command == "recipe":
            # make directory?
            if not (d := Path("./dist-conda")).exists():
                d.mkdir()

            run(
                "grayskull",
                "pypi",
                sdist_path,
                "--sections",
                "package",
                "source",
                "build",
                "requirements",
                "-o",
                "dist-conda",
            )

            _append_recipe(
                f"dist-conda/{PACKAGE_NAME}/meta.yaml",
                "config/recipe-append.yaml",
            )
            session.run("cat", f"dist-conda/{PACKAGE_NAME}/meta.yaml", external=True)

        elif command == "recipe-full":
            import tempfile

            with tempfile.TemporaryDirectory() as d:  # type: ignore[assignment,unused-ignore]
                run(
                    "grayskull",
                    "pypi",
                    sdist_path,
                    "-o",
                    str(d),
                )
                path = Path(d) / PACKAGE_NAME / "meta.yaml"
                session.log(f"cat {path}:")
                with path.open() as f:
                    for line in f:
                        print(line, end="")  # noqa: T201


@nox.session(name="conda-build", **CONDA_DEFAULT_KWS)
@add_opts
def conda_build(session: nox.Session, opts: SessionParams) -> None:
    """Run `conda mambabuild`."""
    runner = Installer.from_envname(
        session=session,
        update=opts.update,
        conda_deps=["boa", "anaconda-client"],
    ).install_all(log_session=opts.log_session)

    cmds, run = opts.conda_build, opts.conda_build_run

    runner.run_commands(run)

    if not run and not cmds:
        cmds = ["build", "clean"]

    if cmds is None:
        cmds = []

    cmds = list(cmds)
    if "clean" in cmds:
        cmds.remove("clean")
        session.log("removing directory dist-conda/build")
        if Path("./dist-conda/build").exists():
            shutil.rmtree("./dist-conda/build")

    for cmd in cmds:
        if cmd == "build":
            if not (d := Path(f"./dist-conda/{PACKAGE_NAME}/meta.yaml")).exists():
                msg = f"no file {d}"
                raise ValueError(msg)

            session.run(
                "conda",
                "mambabuild",
                "--output-folder=dist-conda/build",
                "--no-anaconda-upload",
                "dist-conda",
            )


# ** Other utilities
@nox.session
@add_opts
def cog(session: nox.Session, opts: SessionParams) -> None:
    """Run cog."""
    Installer.from_envname(
        session=session,
        update=opts.update,
        pip_deps="cogapp",
        package="-e .",
    ).install_all(log_session=opts.log_session)
    session.run("cog", "-rP", "README.md", env={"COLUMNS": "100"})


# * Utilities -------------------------------------------------------------------------
def _create_doc_examples_symlinks(session: nox.Session, clean: bool = True) -> None:  # noqa: C901
    """Create symlinks from docs/examples/*.md files to /examples/usage/..."""
    import os

    def usage_paths(path: Path) -> Iterator[Path]:
        with path.open("r") as f:
            for line in f:
                if line.startswith("usage/"):
                    yield Path(line.strip())

    def get_target_path(
        usage_path: str | Path,
        prefix_dir: str | Path = "./examples",
        exts: Sequence[str] = (".md", ".ipynb"),
    ) -> Path:
        path = Path(prefix_dir) / Path(usage_path)

        if not all(ext.startswith(".") for ext in exts):
            msg = "Bad extensions.  Should start with '.'"
            raise ValueError(msg)

        if path.exists():
            return path

        for ext in exts:
            p = path.with_suffix(ext)
            if p.exists():
                return p

        msg = f"no path found for base {path}"
        raise ValueError(msg)

    root = Path("./docs/examples/")
    if clean:
        import shutil

        shutil.rmtree(root / "usage", ignore_errors=True)

    # get all md files
    paths = list(root.glob("*.md"))

    # read usage lines
    for path in paths:
        for usage_path in usage_paths(path):
            target = get_target_path(usage_path)
            link = root / usage_path.parent / target.name

            if link.exists():
                link.unlink()

            link.parent.mkdir(parents=True, exist_ok=True)

            target_rel = os.path.relpath(target, start=link.parent)
            session.log(f"linking {target_rel} -> {link}")

            os.symlink(target_rel, link)


def _append_recipe(recipe_path: str | Path, append_path: str | Path) -> None:
    recipe_path = Path(recipe_path)
    append_path = Path(append_path)

    with recipe_path.open() as f:
        recipe = f.readlines()

    with append_path.open() as f:
        append = f.readlines()

    with recipe_path.open("w") as f:
        f.writelines([*recipe, "\n", *append])
